/**
	Echo Module - To provide an echo of upto 1 second
	\author Mohammad Tariq Rafique
	\date March 1st, 2005
	\Revision 1.0

	Requires the following inputs
	fsl_m_dataout
	fsl_s_datain
	fsl_m_write // ECHO Ready to write out data
	fsl_s_exists // ECHO Should read in data from input
	fsl_m_read
	fsl_m_full // Output buffer is full


*/
module echo_block(clk, Reset, fsl_s_clk, fsl_s_datain, fsl_s_exists, 
				fsl_m_full,	fsl_s_read, fsl_s_control, fsl_m_clk, 
				fsl_m_dataout, fsl_m_write, fsl_m_control);

	parameter ADDR_SIZE=14;

	input clk;
	input Reset;
	
	// FSL Communication I/O
	input [15:0] fsl_s_datain;
	input fsl_s_exists, fsl_m_full;
	input fsl_s_control;
	output [15:0] fsl_m_dataout;
	output fsl_m_write, fsl_s_read;
	output fsl_m_control, fsl_m_clk, fsl_s_clk;
	// Declare as reg as well because their value
	// is changed within an 'always@' loop
	reg fsl_m_write, fsl_s_read;

	// State Parameters
	reg [4:1] y;
	parameter [4:1] IDLE=0, WRITE_OUT=1, WRITE_WAIT=2;
	parameter [4:1] FINISH_WRITE=3, READ_BEGIN=4, FSL_WRITE=5;
	parameter [4:1] FSL_WRITE_FINISH=6, FSL_WRITE_WAIT=7;
	parameter [4:1] ADD_DATA=8, ADD_DATA_WAIT=9, RANDOM_WAIT=10;

	reg [(ADDR_SIZE-1):0] read_addr; //Write Data to this point
	reg [(ADDR_SIZE-1):0] write_addr; //Read data from this address
	reg [(ADDR_SIZE-1):0] addr; // Connector to port memory
	reg we; // Write Enable for block memory
	wire [15:0] currDataLink; // Data read from block memory
	// is not really 'current' since it is 16K+ samples behind
	// incoming data
	reg [15:0] currData;
	
	block_memory echo_buffer (
		.addr(addr),
		.clk(clk),
		.din(fsl_s_datain),
		.dout(currDataLink),
		.we(we));

	always @(posedge clk or posedge Reset)
	begin
		if ( Reset==1 )
		begin
			// Start reading from beginning of block
			read_addr = 0;
			// Start writing from end of block.
			write_addr = 16383;
			we = 0;
			y <= IDLE;
		end
		else
		begin
			case (y)
				IDLE:// Wait, just do nothing
				begin
					fsl_s_read = 0;
					assign we = 0;
					if (fsl_s_exists) // Check for FSL Data
					begin // receiving data from audio
						addr = write_addr;
						y <= WRITE_OUT;
					end
					else
					begin
						y <= IDLE;
					end
				end
				WRITE_OUT:
				begin
					we = 1;
					y <= WRITE_WAIT;
				end
				WRITE_WAIT: // Ensure write is completed
				begin
					y <= FINISH_WRITE;
				end
				FINISH_WRITE:
				begin
					// Disable write mechanism
					we = 0;
					y <= READ_BEGIN;
				end
				READ_BEGIN: // READ contents from buffer to output to FSL
				begin
					// Since write_addr is only 15 bits, and there
					// are only 16384 blocks, the bits will wrap
					// around when they overflow (circular buffer)
					write_addr = write_addr + 1;
					addr = read_addr;
					y <= RANDOM_WAIT;
				end
				RANDOM_WAIT: // Couldn't think of a state name
				// Wait for read to compelete
					y <= ADD_DATA_WAIT;
				ADD_DATA_WAIT:
				begin  // Add old sample to incoming sample
					// Modify CurrData
					currData = currDataLink; // REad from memory
					
					// Need to attenuate old sample so that 
					// AC97 does not saturate
					// The 'if-else' below will attempt to
					// do a signed right shift
					if (currData[15] == 1) 
					begin
						currData = currData >> 4;
						currData = currData | 16'b1111000000000000;
					end
					else
					begin
						currData = currData >> 4;
					end
					y <= FSL_WRITE_WAIT;
				end
				// Now write out data to FSL bus
				FSL_WRITE_WAIT:
				begin // Ensure data has settled
					y <= FSL_WRITE;
				end
				FSL_WRITE:
				begin
					fsl_m_write = 1;
					y <= FSL_WRITE_FINISH;
				end
				FSL_WRITE_FINISH:
				begin
					fsl_m_write = 0;
					// Acknowledge to FSL, done read
					// Since write has completed, move onto
					// next data on FIFO
					fsl_s_read = 1;
					read_addr = read_addr + 1;
					y <= IDLE; // Go back to beginning
				end
				default:
				begin
					y <= IDLE;
				end
			endcase
		end // if (reset)
	end // @always
	assign fsl_m_dataout = fsl_s_datain + currData;
endmodule