library IEEE;
use IEEE.std_logic_1164.all;
use ieee.numeric_std.all;
use IEEE.std_logic_unsigned.all;

entity FSL_FFT is
  generic (
    C_DWIDTH      : integer := 32
  );
  port (
    -- Global signals
    Clk   : in std_logic;
    Reset : in std_logic;

    -- Slave FSL Signals
    FSL_S_Clk     : out std_logic;
    FSL_S_Read    : out std_logic;
    FSL_S_Data    : in  std_logic_vector(0 to C_DWIDTH-1);
    FSL_S_Control : in  std_logic;
    FSL_S_Exists  : in  std_logic;

    -- Master FSL Signals
    FSL_M_Clk     : out std_logic;
    FSL_M_Write   : out std_logic;
    FSL_M_Data    : out std_logic_vector(0 to C_DWIDTH-1);
    FSL_M_Control : out std_logic;
    FSL_M_Full    : in  std_logic;

    -- External Controller Signals
    Frame_Ready   : in std_logic;
    Frame_Done    : out std_logic;
    blk_exp	   : out std_logic_vector(4 downto 0)
    );

end entity FSL_FFT;

library unisim;
use unisim.all;

architecture IMP of FSL_FFT is

    component fft512
	port (
	xn_re: IN std_logic_VECTOR(15 downto 0);
	xn_im: IN std_logic_VECTOR(15 downto 0);
	start: IN std_logic;
	unload: IN std_logic;
	nfft: IN std_logic_VECTOR(4 downto 0);
	nfft_we: IN std_logic;
	fwd_inv: IN std_logic;
	fwd_inv_we: IN std_logic;
	clk: IN std_logic;
	xk_re: OUT std_logic_VECTOR(15 downto 0);
	xk_im: OUT std_logic_VECTOR(15 downto 0);
	xn_index: OUT std_logic_VECTOR(8 downto 0);
	xk_index: OUT std_logic_VECTOR(8 downto 0);
	rfd: OUT std_logic;
	busy: OUT std_logic;
	dv: OUT std_logic;
	edone: OUT std_logic;
	done: OUT std_logic;
	blk_exp: OUT std_logic_VECTOR(4 downto 0));
    end component;

    signal xn_re: std_logic_vector(15 downto 0);
    signal start: std_logic;
    signal unload: std_logic;
    signal nfft_we: std_logic;
    signal fwd_inv_we: std_logic;
    signal xk_re: std_logic_vector(15 downto 0);
    signal xk_im: std_logic_vector(15 downto 0);
    signal xn_index: std_logic_vector(8 downto 0);
    signal xk_index: std_logic_vector(8 downto 0);
    signal rfd: std_logic;
    signal busy: std_logic;
    signal dv: std_logic;
    signal edone: std_logic;
    signal done: std_logic;
    --signal blk_exp: std_logic_vector(4 downto 0);
    signal nCount: std_logic_vector(8 downto 0);
    signal nOutWait: std_logic_vector(2 downto 0);

    type state is (INIT, IDLE, DOFFT, READ, WRITE_WAIT, UNLOAD_WAIT, WRITE);
    signal op: state;

begin  -- architecture IMP

    -- instantiate the FFT core
    fft512_instance : fft512
		port map (
			xn_re => xn_re,
			xn_im => "0000000000000000",
			start => start,
            unload => unload,
            nfft => "01001",
            nfft_we => nfft_we,
			fwd_inv => '1',
			fwd_inv_we => fwd_inv_we,
			clk => clk,
			xk_re => xk_re,
			xk_im => xk_im,
			xn_index => xn_index,
			xk_index => xk_index,
			rfd => rfd,
			busy => busy,
			dv => dv,
			edone => edone,
			done => done,
            blk_exp => blk_exp);

    process(Clk, Reset)
    begin
        if (Reset = '1') then
            op <= INIT;
        elsif (Clk'event and Clk = '1') then
            case op is
                -- initialize variables
                when INIT =>
                    FSL_S_Read <= '0';
                    FSL_M_Write <= '0';
                    FSL_M_Data <= "00000000000000000000000000000000";
                    FSL_M_Control <= '0';
                    op <= IDLE;
                    start <= '0';
                    nfft_we <= '1';
                    fwd_inv_we <= '1';
                    nCount <= "000000000";
                    nOutWait <= "000";
                    unload <= '0';
                -- wait for data and control start
                when IDLE =>
                    FSL_S_Read <= '0';
                    FSL_M_Write <= '0';
                    Frame_Done <= '0';
                    nfft_we <= '0';
                    fwd_inv_we <= '0';
                    if (FSL_S_Exists = '1' and Frame_Ready = '1') then
                        op <= DOFFT;
                        start <= '1';
                    else
                        op <= IDLE;
                    end if;
                -- this used to wait until the appropriate index value
                -- to start sending data in, but it now just acts as a
                -- 1 cycle delay
                when DOFFT =>
                    start <= '0';
                    if (rfd = '1' and xn_index = "000000000") then
                        op <= READ;
                    else
                        op <= DOFFT;
                    end if;
                -- read data from input FIFO and send to FFT core
                when READ =>
                    FSL_S_Read <= '1';
                    xn_re <= FSL_S_Data(16 to 31);
                    if (nCount = "111111111") then
                        op <= WRITE_WAIT;
                        nCount <= "000000000";
                    else
                        nCount <= nCount + 1;
                        op <= READ;
                    end if;
                -- wait until FFT core has processed data
                when WRITE_WAIT =>
                    FSL_S_Read <= '0';
                    if (done = '1') then
                        op <= UNLOAD_WAIT;
                        unload <= '1';
                    else
                        op <= WRITE_WAIT;
                    end if;
                -- signal the FFT core to unload in natural order
                when UNLOAD_WAIT =>
                    unload <= '0';
                    if (nOutWait = "100") then
                        op <= WRITE;
                        nOutWait <= "000";
                    else
                        op <= UNLOAD_WAIT;
                        nOutWait <= nOutWait + 1;
                    end if;
                -- write out FFT data to output FIFO
                when WRITE =>
                    FSL_M_Write <= '1';
                    FSL_M_Data(0 to 15) <= xk_im;
                    FSL_M_Data(16 to 31) <= xk_re;
                    if (nCount = "111111111") then
                        op <= IDLE;
                        nCount <= "000000000";
                        Frame_Done <= '1';
                    else
                        nCount <= nCount + 1;
                        op <= WRITE;
                    end if;
            end case;
        end if;
    end process;

end architecture IMP;
