library IEEE;
use IEEE.std_logic_1164.all;
use ieee.numeric_std.all;
use IEEE.std_logic_signed.all;

entity FFT_CTRL is
  generic (
    C_DWIDTH      : integer := 32
  );
  port (
    -- Global signals
    Clk   : in std_logic;
    Reset : in std_logic;

    -- Slave FSL Signals
    FSL_S_Clk     : out std_logic;
    FSL_S_Read    : out std_logic;
    FSL_S_Data    : in  std_logic_vector(0 to C_DWIDTH-1);
    FSL_S_Control : in  std_logic;
    FSL_S_Exists  : in  std_logic;

    -- Master FSL Signals
    FSL_M_Clk     : out std_logic;
    FSL_M_Write   : out std_logic;
    FSL_M_Data    : out std_logic_vector(0 to C_DWIDTH-1);
    FSL_M_Control : out std_logic;
    FSL_M_Full    : in  std_logic;

    -- External Controller Signals
    Frame_Done    : out std_logic;
    Scale_Ready   : in std_logic;
    blk_exp	   : in std_logic_vector(4 downto 0)
    );

end entity FFT_CTRL;

library unisim;
use unisim.all;

architecture IMP of FFT_CTRL is

    type state_in is (INIT_I, IDLE_I, START_FFT_I);
    signal op_in: state_in;

    type state_out is (INIT_O, IDLE_O, WRITE_O);
    signal op_out: state_out;

begin  -- architecture IMP

    -- this process handles start control
    process(Clk, Reset)
    begin
        if (Reset = '1') then
            op_in <= INIT_I;
        elsif (Clk'event and Clk = '1') then
            case op_in is
                -- initialize variables
                when INIT_I =>
                    FSL_S_Read <= '0';
                    op_in <= IDLE_I;
                -- wait until data available in input
                -- FIFO, then signal control signal
                when IDLE_I =>
                    FSL_S_Read <= '0';
                    if (FSL_S_Exists = '1') then
                        op_in <= START_FFT_I;
                        Frame_Done <= '1';
                    else
                        op_in <= IDLE_I;
                    end if;
                -- pop the data off the FIFO and wait
                when START_FFT_I =>
                    FSL_S_Read <= '1';
                    Frame_Done <= '0';
                    op_in <= IDLE_I;
            end case;
        end if;
    end process;

    -- this process handles scale factor write back
    process(Clk, Reset)
    begin
        if (Reset = '1') then
            op_out <= INIT_O;
        elsif (Clk'event and Clk = '1') then
            case op_out is
                -- initialize variables
                when INIT_O =>
                    FSL_M_Write <= '0';
                    FSL_M_Data <= "00000000000000000000000000000000";
                    FSL_M_Control <= '0';
                    op_out <= IDLE_O;
                -- if done control signalled then write out
                -- scale factor to output FIFO
                when IDLE_O =>
                    if (Scale_Ready = '1') then
				    FSL_M_Data(27 to 31) <= blk_exp;
				    FSL_M_Write <= '1';
                        op_out <= WRITE_O;
                    else
                        op_out <= IDLE_O;
                    end if;
                when WRITE_O =>
                    FSL_M_Write <= '0';
                    op_out <= IDLE_O;
            end case;
        end if;
    end process;

end architecture IMP;
