library IEEE;
use IEEE.std_logic_1164.all;
use ieee.numeric_std.all;
use IEEE.std_logic_unsigned.all;

entity FSL_INBUF is
  generic (
    C_DWIDTH      : integer := 32
  );
  port (
    -- Global signals
    Clk   : in std_logic;
    Reset : in std_logic;

    -- Slave FSL Signals
    FSL_S_Clk     : out std_logic;
    FSL_S_Read    : out std_logic;
    FSL_S_Data    : in  std_logic_vector(0 to C_DWIDTH-1);
    FSL_S_Control : in  std_logic;
    FSL_S_Exists  : in  std_logic;

    -- Master FSL Signals
    FSL_M_Clk     : out std_logic;
    FSL_M_Write   : out std_logic;
    FSL_M_Data    : out std_logic_vector(0 to C_DWIDTH-1);
    FSL_M_Control : out std_logic;
    FSL_M_Full    : in  std_logic;

    -- External Control Signals
    Frame_Done    : out std_logic
    );

end entity FSL_INBUF;

library unisim;
use unisim.all;

architecture IMP of FSL_INBUF is

    component inbuf
    port (
        addr: IN std_logic_VECTOR(8 downto 0);
        clk: IN std_logic;
        din: IN std_logic_VECTOR(15 downto 0);
        dout: OUT std_logic_VECTOR(15 downto 0);
        nd: IN std_logic;
        rfd: OUT std_logic;
        rdy: OUT std_logic;
        we: IN std_logic);
    end component;

    signal nAddr: std_logic_vector(8 downto 0);
    signal nDIn: std_logic_vector(15 downto 0);
    signal nDOut: std_logic_vector(15 downto 0);
    signal we: std_logic;
    signal nd: std_logic;
    signal rfd: std_logic;
    signal rdy: std_logic;
    signal nCount: std_logic_vector(9 downto 0);

    type state is (INIT, IDLE, READ, READ_ACK, WRITE, WRITE_ACK, WRITE_WAIT);
    signal op: state;

begin  -- architecture IMP

    -- instantiate internal buffer
    in_buffer : inbuf
    port map (
        addr => nAddr,
        clk => clk,
        din => nDIn,
        dout => nDOut,
        nd => nd,
        rfd => rfd,
        rdy => rdy,
        we => we);

    process(Clk, Reset)
    begin
        if (Reset = '1') then
            op <= INIT;
        elsif (Clk'event and Clk = '1') then
            case op is
                -- initialize variables
                when INIT =>
                    FSL_S_Read <= '0';
                    FSL_M_Write <= '0';
                    FSL_M_Data <= "00000000000000000000000000000000";
                    FSL_M_Control <= '0';
                    op <= IDLE;
                    nAddr <= "000000000";
                    nDIn <= "0000000000000000";
                    we <= '1';
                    nd <= '0';
                    nCount <= "0000000000";
                -- wait for data in input FIFO and RAM to be ready for data
                when IDLE =>
                    FSL_S_Read <= '0';
                    FSL_M_Write <= '0';
                    Frame_Done <= '0';
                    if (FSL_S_Exists = '1' and rfd = '1') then
                        op <= READ;
                        nDIn <= FSL_S_Data(16 to 31);
                        nd <= '1';
                    else
                        op <= IDLE;
                    end if;
                -- pop read data off input FIFO and increment pointers
                when READ =>
                    nd <= '0';
                    FSL_S_Read <= '1';
                    nAddr <= nAddr + 1;
                    nCount <= nCount + 1;
                    op <= READ_ACK;
                -- check if buffer is full, otherwise keep reading in
                -- if full, go on to write contents of buffer out
                when READ_ACK =>
                    FSL_S_Read <= '0';
                    if (nCount = "1000000000") then
                        nAddr <= "000000000";
                        nCount <= "0000000000";
                        we <= '0';
                        op <= WRITE;
                        nd <= '1';
                    else
                        op <= IDLE;
                    end if;
                -- wait until data has been read from buffer, then
                -- write to output FIFO
                when WRITE =>
                    nd <= '0';
                    FSL_M_Write <= '0';
                    if (FSL_M_Full = '0') then
                        if (rdy = '1') then
                            op <= WRITE_WAIT;
                            FSL_M_Data(16 to 31) <= nDOut;
                        end if;
                    else
                        op <= WRITE;
                    end if;
                -- increment pointers and push data to output FIFO
                when WRITE_WAIT =>
                    nAddr <= nAddr + 1;
                    nCount <= nCount + 1;
                    op <= WRITE_ACK;
                    FSL_M_Write <= '1';
                    if (nCount /= "0000000001") then
                        nd <= '1';
                    end if;
                -- check if buffer emptied out
                when WRITE_ACK =>
                    nd <= '0';
                    FSL_M_Write <= '0';
                    --if (nCount = "1000000000") then
                    if (nCount = "0000000010") then
                        nAddr <= "000000000";
                        nCount <= "0000000000";
                        we <= '1';
                        op <= IDLE;
                        Frame_Done <= '1';
                    else
                        op <= WRITE;
                    end if;
            end case;
        end if;
    end process;

end architecture IMP;
