library IEEE;
use IEEE.std_logic_1164.all;
use ieee.numeric_std.all;
use IEEE.std_logic_unsigned.all;

entity FSL_EQ is
  generic (
    C_DWIDTH      : integer := 32
  );
  port (
    -- Global signals
    Clk   : in std_logic;
    Reset : in std_logic;

    -- Slave FSL Signals
    FSL_S_Clk     : out std_logic;
    FSL_S_Read    : out std_logic;
    FSL_S_Data    : in  std_logic_vector(0 to C_DWIDTH-1);
    FSL_S_Control : in  std_logic;
    FSL_S_Exists  : in  std_logic;

    -- Master FSL Signals
    FSL_M_Clk     : out std_logic;
    FSL_M_Write   : out std_logic;
    FSL_M_Data    : out std_logic_vector(0 to C_DWIDTH-1);
    FSL_M_Control : out std_logic;
    FSL_M_Full    : in  std_logic;

    -- Control Signal to Start IFFT
    Frame_Done    : out std_logic
    );

end entity FSL_EQ;

library unisim;
use unisim.all;

architecture IMP of FSL_EQ is

    component gains
    port (
        addra: IN std_logic_VECTOR(8 downto 0);
        addrb: IN std_logic_VECTOR(8 downto 0);
        clka: IN std_logic;
        clkb: IN std_logic;
        dinb: IN std_logic_VECTOR(15 downto 0);
        douta: OUT std_logic_VECTOR(15 downto 0);
        nda: IN std_logic;
        ndb: IN std_logic;
        rfda: OUT std_logic;
        rfdb: OUT std_logic;
        rdya: OUT std_logic;
        rdyb: OUT std_logic;
        web: IN std_logic);
    end component;

    component mult
    port (
        clk: IN std_logic;
        a: IN std_logic_VECTOR(15 downto 0);
        b: IN std_logic_VECTOR(15 downto 0);
        q: OUT std_logic_VECTOR(31 downto 0);
        rfd: OUT std_logic;
        nd: IN std_logic;
        rdy: OUT std_logic);
    end component;

    signal nWAddr: std_logic_vector(8 downto 0);
    signal nRAddr: std_logic_vector(8 downto 0);
    signal nDIn: std_logic_vector(15 downto 0);
    signal nDOut: std_logic_vector(15 downto 0);
    signal nda: std_logic;
    signal ndb: std_logic;
    signal rfda: std_logic;
    signal rfdb: std_logic;
    signal rdya: std_logic;
    signal rdyb: std_logic;
    signal web: std_logic;

    signal nMultInR: std_logic_vector(15 downto 0);
    signal nMultRq: std_logic_vector(31 downto 0);
    signal rfd_rm: std_logic;
    signal nd_rm: std_logic;
    signal rdy_rm: std_logic;

    signal nMultInI: std_logic_vector(15 downto 0);
    signal nMultIq: std_logic_vector(31 downto 0);
    signal rfd_im: std_logic;
    signal nd_im: std_logic;
    signal rdy_im: std_logic;

    signal nGain: std_logic_vector(15 downto 0);
    signal nTemp1: std_logic_vector(31 downto 0);
    signal nTemp2: std_logic_vector(31 downto 0);

    type state is (INIT, GETGAIN, IDLE, READ, MULTIPLY, WRITE, OUTPUT, WRITE_ACK);
    signal op: state;
  
begin  -- architecture IMP

    -- instantiate RAM for gains
    gains_instance : gains
    port map (
        addra => nRAddr,
        addrb => nWAddr,
        clka => clk,
        clkb => clk,
        dinb => nDIn,
        douta => nDOut,
        nda => nda,
        ndb => ndb,
        rfda => rfda,
        rfdb => rfdb,
        rdya => rdya,
        rdyb => rdyb,
        web => web);

    -- instantiate 2 multipliers for real and imaginary
    multr_instance : mult
    port map (
        clk => clk,
        a => nMultInR,
        b => nGain,
        q => nMultRq,
        rfd => rfd_rm,
        nd => nd_rm,
        rdy => rdy_rm);

    multi_instance : mult
    port map (
        clk => clk,
        a => nMultInI,
        b => nGain,
        q => nMultIq,
        rfd => rfd_im,
        nd => nd_im,
        rdy => rdy_im);

    process(Clk, Reset)
    begin
        if (Reset = '1') then
            op <= INIT;
        elsif (Clk'event and Clk = '1') then
            case op is
                -- initialize all variables
                when INIT =>
                    FSL_S_Read <= '0';
                    FSL_M_Write <= '0';
                    FSL_M_Data <= "00000000000000000000000000000000";
                    FSL_M_Control <= '0';
                    op <= GETGAIN;
                    nWAddr <= "000000000";
                    nRAddr <= "000000000";
                    nDIn <= "0000000000000000";
                    nda <= '1';
                    ndb <= '0';
                    web <= '0';
                    nMultInR <= "0000000000000000";
                    nd_rm <= '0';
                    nMultInI <= "0000000000000000";
                    nd_im <= '0';
                    Frame_Done <= '0';
                -- get the gain for the current index
                when GETGAIN =>
                    FSL_M_Write <= '0';
                    Frame_Done <= '0';
                    nda <= '0';
                    -- check that read has completed
                    if (rdya = '1') then
                        nGain <= nDOut;
                        op <= IDLE;
                    -- otherwise wait for read to finish
                    else
                        op <= GETGAIN;
                    end if;
                -- wait for data on input FIFO
                when IDLE =>
                    if (FSL_S_Exists = '1') then
                        op <= READ;
                        nMultInI <= FSL_S_Data(0 to 15);
                        nMultInR <= FSL_S_Data(16 to 31);
                    else
                        op <= IDLE;
                    end if;
                -- signal FSL to pop data off
                when READ =>
                    FSL_S_Read <= '1';
                    nRAddr <= nRAddr + 1;
                    op <= MULTIPLY;
                    nda <= '1';
                -- begin the multiplication
                when MULTIPLY =>
                    FSL_S_Read <= '0';
                    nda <= '0';
                    if (rfd_rm = '1' and rfd_im = '1') then
                        nd_rm <= '1';
                        nd_im <= '1';
                        op <= WRITE;
                    else
                        op <= MULTIPLY;
                    end if;
                -- once multiplication done, check for overflow
                when WRITE =>
                    nd_rm <= '0';
                    nd_im <= '0';
                    if (rdy_rm = '1' and rdy_im = '1' and FSL_M_Full = '0') then
                        -- check for negative overflow
                        if (nMultRq(31) = '1') then
                            if (nMultRq(31 downto 16) = "1111111111111111") then
                                nTemp1 <= std_logic_vector((signed(nMultRq) srl 1));
                            else
                                nTemp1 <= "00000000000000001000000000000000";
                            end if;
                        -- check for positive overflow
                        else
                            if (nMultRq(31 downto 16) = "0000000000000000") then
                                nTemp1 <= std_logic_vector((signed(nMultRq) srl 1));
                            else
                                nTemp1 <= "00000000000000000111111111111111";
                            end if; 
                        end if;
                        -- check for negative overflow
                        if (nMultIq(31) = '1') then
                            if (nMultIq(31 downto 16) = "1111111111111111") then
                                nTemp2 <= std_logic_vector((signed(nMultIq) srl 1));
                            else
                                nTemp2 <= "00000000000000001000000000000000";
                            end if;
                        -- check for positive overflow
                        else
                            if (nMultIq(31 downto 16) = "0000000000000000") then
                                nTemp2 <= std_logic_vector((signed(nMultIq) srl 1));
                            else
                                nTemp2 <= "00000000000000000111111111111111";
                            end if; 
                        end if;
                        op <= OUTPUT;
                    else
                        op <= WRITE;
                    end if;
                -- push processed data onto output FIFO
                when OUTPUT =>
                    FSL_M_Data(0 to 15) <= nTemp2(15 downto 0);
                    FSL_M_Data(16 to 31) <= nTemp1(15 downto 0);
                    op <= WRITE_ACK;
                -- begin read of new gain value
                when WRITE_ACK =>
                    op <= GETGAIN;
                    FSL_M_Write <= '1';
                    if (nRAddr = "111111111") then
                        Frame_Done <= '1';
                        nRAddr <= "000000000";
                    end if;
                    nda <= '1';
            end case;
        end if;
    end process;

end architecture IMP;
